# -*- coding: utf-8 -*-
import collections

from ..operator import Operator


class KeysOperator(Operator):
    """
    Asserts that a given dictionary has a key or keys.

    Example::

        # Should style
        {'foo': True} | should.have.key('foo')
        {'foo': True, 'bar': False} | should.have.keys('bar', 'foo')

        # Should style - negation form
        {'bar': True} | should.not_have.key('foo')
        {'baz': True, 'bar': False} | should.not_have.keys('bar', 'foo')

        # Expect style
        {'foo': True} | expect.to.have.key('foo')
        {'foo': True, 'bar': False} | expect.to.have.keys('bar', 'foo')

        # Expect style - negation form
        {'bar': True} | expect.to_not.have.key('foo')
        {'baz': True, 'bar': False} | expect.to_not.have.keys('bar', 'foo')
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('keys', 'key',)

    # Operator chain aliases
    aliases = ('present', 'equal', 'to')

    # Expected message templates
    expected_message = Operator.Dsl.Message(
        'a dictionary-like object that has the key(s) "{value}"',
        'a dictionary-like object that has not the key(s) "{value}"',
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with value "{value}"',
    )

    def after_success(self, obj, *keys):
        if not self.ctx.negate:
            self.ctx.value = [obj[x] for x in obj if x in keys]

        if len(keys) == 1:
            self.ctx.value = self.ctx.value[0]

    def match(self, subject, *keys, **kw):
        if self._not_a_dict(subject):
            return False, ['subject is not a dict type']
        return self._matches(subject, (keys, kw))

    def _not_a_dict(self, value):
        return not isinstance(value, collections.Mapping)

    def _matches(self, subject, expected):
        args, kwargs = expected

        reasons = []
        for name in args:
            has_key, reason = self._has_key(subject, name)
            if not has_key:
                return False, [reason]
            else:
                reasons.append(reason)

        for name, value in kwargs.items():
            has_key, reason = self._has_key(subject, name, value)
            if not has_key:
                return False, [reason]
            else:
                reasons.append(reason)

        return True, reasons

    def _has_key(self, subject, name, *args):
        if args:
            expected_value = args[0]

            try:
                value = subject[name]
            except KeyError:
                return False, 'key {0!r} {1!r} not found'.format(
                    name, expected_value)
            else:
                result, _ = expected_value._match(value)
                reason_message = 'not found' if not result else 'found'
                return result, 'key {0!r} {1!r} {2}'.format(
                    name, expected_value, reason_message)

        if name in subject:
            return True, 'key {0!r} found'.format(name)

        return False, 'key {0!r} not found'.format(name)



# -*- coding: utf-8 -*-
from ..operator import Operator


class PropertyOperator(Operator):
    """
    Asserts if a given object has property or properties.

    Example::

        class Foo(object):
            bar = 'foo'

            def baz():
                pass

        # Should style
        Foo() | should.have.property('bar')
        Foo() | should.have.properties('bar', 'baz')
        Foo() | should.have.properties.present.equal.to('bar', 'baz')

        # Should style - negation form
        Foo() | expect.to_not.have.property('bar')
        Foo() | expect.to_not.have.properties('bar', 'baz')
        Foo() | expect.to_not.have.properties.present.equal.to('bar', 'baz')

        # Expect style
        Foo() | expect.to.have.property('bar')
        Foo() | expect.to.have.properties('bar', 'baz')
        Foo() | expect.to.have.properties.present.equal.to('bar', 'baz')

        # Expect style - negation form
        Foo() | expect.to_not.have.property('bar')
        Foo() | expect.to_not.have.properties('bar', 'baz')
        Foo() | expect.to_not.have.properties.present.equal.to('bar', 'baz')
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('properties', 'property', 'attribute', 'attributes')

    # Operator chain aliases
    aliases = ('present', 'equal', 'to')

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'an object that has the following properties "{value}"',
        'an object that has not the following properties "{value}"',
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with data "{value}"',
    )

    def after_success(self, obj, *keys):
        if self.ctx.negate:
            return

        # Get attribute keys
        self.ctx.value = [getattr(obj, x) for x in keys]

        if len(keys) == 1:
            self.ctx.value = self.ctx.value[0]

    def match(self, subject, *args, **kwargs):
        success_reasons = []
        for name in args:
            has_property, reason = self._has_property(subject, name)
            if not has_property:
                return False, [reason]
            else:
                success_reasons.append(reason)

        for name, value in kwargs.items():
            has_property, reason = self._has_property(subject, name, value)
            if not has_property:
                return False, [reason]
            else:
                success_reasons.append(reason)

        return True, success_reasons

    def _has_property(self, subject, name, *args):
        if args:
            try:
                value = getattr(subject, name)
            except AttributeError:
                return False, 'property {0!r} not found'.format(name)
            else:
                expected_value = args[0]
                result, _ = expected_value._match(value)
                if not result:
                    return False, 'property {0!r} {1!r} not found'.format(
                        name, expected_value)
                return True, 'property {0!r} {1!r} found'.format(
                    name, expected_value)

        if not hasattr(subject, name):
            return False, 'property {0!r} not found'.format(name)
        return True, 'property {0!r} found'.format(name)



import pytest
from grappa.test import Test as GrappaTest
from grappa import should, expect


@pytest.mark.parametrize('operator', (
    'to', 'has', 'have', 'include', 'do'
))
def test_assertion_attributes(operator):
    assert isinstance(getattr(should, operator), GrappaTest)
    assert getattr(should, operator)._ctx.negate is False
    assert isinstance(getattr(expect, operator), GrappaTest)
    assert getattr(expect, operator)._ctx.negate is False


@pytest.mark.parametrize('operator', (
    'which', 'that'
))
def test_chain_attributes(operator):
    assert isinstance(getattr(should, operator), GrappaTest)
    assert getattr(expect, operator)._ctx.reset is True
    assert getattr(expect, operator)._ctx.negate is False
    assert isinstance(getattr(expect, operator), GrappaTest)
    assert getattr(expect, operator)._ctx.reset is True
    assert getattr(expect, operator)._ctx.negate is False


@pytest.mark.parametrize('operator', (
    'not_to', 'to_not', 'does_not', 'do_not', '_not',
    'not_have', 'not_has', 'have_not', 'has_not', 'dont'
))
def test_negation_attributes(operator):
    assert isinstance(getattr(should, operator), GrappaTest)
    assert getattr(should, operator)._ctx.negate
    assert isinstance(getattr(expect, operator), GrappaTest)
    assert getattr(expect, operator)._ctx.negate



import pytest


def test_expect_keys(should):
    {'foo': 'bar'} | should.have.key('foo')
    {'foo': 'bar'} | should.have.key('foo') > should.be.equal.to('bar')
    'bar' | should.be.equal.to('bar')

    should({'foo': 'bar'}).have.key('foo') > should.be.equal.to('bar')
    should({'foo': 'bar'}).have.key('foo').which.should.be.equal.to('bar')

    with pytest.raises(AssertionError):
        {'foo': 'bar'} | should.have.key('bar')

    with pytest.raises(AssertionError):
        [] | should.have.key('bar')

    with pytest.raises(AssertionError):
        should({'foo': 'bar'}).have.key('foo') > should.be.equal.to('foo')

    with pytest.raises(AssertionError):
        should({'foo': 'bar'}).have.key('foo').which.should.be.equal.to('pepe')



# -*- coding: utf-8 -*-
from ..operator import Operator


class EqualOperator(Operator):
    """
    Performs a strict equality comparison between ``x`` and ``y`` values.

    Uses ``==`` built-in binary operator for the comparison.

    Example::

        # Should style
        'foo' | should.be.equal('foo')
        'foo' | should.be.equal.to('foo')
        'foo' | should.be.equal.to.value('foo')

        # Should style - negation form
        'foo' | should.not_be.equal('foo')
        'foo' | should.not_be.equal.to('foo')
        'foo' | should.not_be.equal.to.value('foo')

        # Expect style
        'foo' | expect.to.equal('foo')
        'foo' | expect.to.equal.to('foo')
        'foo' | expect.to.equal.to.value('foo')

        # Expect style - negation form
        'foo' | expect.to_not.equal('foo')
        'foo' | expect.to_not.equal.to('foo')
        'foo' | expect.to_not.equal.to.value('foo')
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('equal', 'same')

    # Operator chain aliases
    aliases = ('value', 'data', 'to', 'of', 'as')

    # Error message templates
    expected_message = Operator.Dsl.Message(
        'a value that is equal to "{value}"',
        'a value that is not equal to "{value}"',
    )

    # Subject message template
    subject_message = Operator.Dsl.Message(
        'an value of type "{type}" with data "{value}"',
    )

    def match(self, subject, expected):
        return subject == expected, []



import pytest


def test_expect_properties(should):
    class foo(object):
        foo = 'bar'

        def bar(self):
            pass

    foo() | should.have.property('foo')
    foo() | should.have.property('foo') > should.be.equal.to('bar')

    should(foo()).have.property('foo') > should.be.equal.to('bar')
    should(foo()).have.property('foo').which.should.be.equal.to('bar')

    should(foo()).have.property('bar').which.should.be.a('method')

    should(foo()).have.properties('bar', 'foo')

    with pytest.raises(AssertionError):
        should(foo()).have.property('baz')

    with pytest.raises(AssertionError):
        should(foo()).have.property('baz')

    with pytest.raises(AssertionError):
        should(foo()).have.property('foo') > should.be.equal.to('foo')

    with pytest.raises(AssertionError):
        should(foo()).have.property('foo').which.should.be.equal.to('pepe')



# -*- coding: utf-8 -*-
from .base import BaseTest


class AssertionProxy(BaseTest):
    """
    AssertionProxy is used as proxy delegator call chain for test
    definitions.
    """

    def __init__(self, test, operator, fn):
        self._fn = fn
        self._test = test
        self._op = operator

    def __call__(self, expected, *args, **kw):
        return self._fn(expected, *args, **kw)

    def __getattr__(self, name):
        # Match chain aliases
        if hasattr(self._op, 'aliases') and name in self._op.aliases:
            self._test._engine.add_keyword(name)
            return self

        # Fallback to parent assertion object
        return getattr(self._test, name)

    def __ror__(self, value):
        """
        Overloads ``|`` operator.
        """
        return self._fn(value)._trigger()

    def __gt__(self, value):
        """
        Overloads ``>`` operator.
        """
        return self.__ror_(value)



# -*- coding: utf-8 -*-
from ..decorators import attribute


@attribute(operators=('to', 'has', 'have', 'include', 'do'))
def be(ctx):
    """
    Semantic attributes providing chainable declarative DSL
    for assertions.
    """
    ctx.negate = False


@attribute(operators=('that',))
def which(ctx):
    """
    Semantic attributes providing chainable declarative DSL
    for assertions chaning.
    """
    ctx.reset = True


@attribute(operators=(
    'not_to', 'to_not', 'does_not', 'do_not', '_not',
    'not_have', 'not_has', 'have_not', 'has_not', 'dont'
))
def not_be(ctx):
    """
    Semantic negation attributes providing chainable declarative DSL
    for assertions.
    """
    ctx.negate = True



# -*- coding: utf-8 -*-
import functools
from colorama import Fore, Style

from .config import config


class ErrorTemplate(object):
    """
    ErrorTemplate is used to compose and render a human-friendly
    descriptive error message.
    """

    separator = ' ' * 6

    section_separator = '\n\n  '

    header = 'Oops! Something went wrong!'

    def __init__(self):
        self.sections = []

    @property
    def color(self):
        return Style.BRIGHT + Fore.GREEN if config.use_colors else ''

    @property
    def reset(self):
        return Style.RESET_ALL if config.use_colors else ''

    def map_iterable(self, content):
        margin = ' ' * 4
        if type(content[0]) is str:
            return ('\n' + margin).join(
                '> ' + item for item in content if item
            )

        return ('\n\n' + margin).join(
            item.render(ErrorTemplate.separator)
            for item in content
            if hasattr(item, 'render')
        )

    def block(self, title, content):
        # If no content, just exit
        if not content:
            return self

        # If iterable type, aggregate elements and format them
        if isinstance(content, (tuple, list)):
            content = self.map_iterable(content)

        # Register section object
        self.sections.append({
            'title': title,
            'content': content
        })

        return self

    def render(self):
        def add_section(buf, section):
            buf.append('{color}{title}{reset}\n    {content}'.format(
                color=self.color,
                title=section['title'],
                reset=self.reset,
                content=section['content']
            ))
            return buf

        sections = functools.reduce(add_section, self.sections, [self.header])
        return self.section_separator.join(sections)



import pytest


def test_should_type(should):
    'foo' | should.be.a(str)
    4 | should.be.an(int)
    {} | should.be.an(dict)
    tuple() | should.be.an(tuple)
    tuple() | should.be.instance.of('tuple')
    (lambda: True) | should.be.instance.of('lambda')

    def foo(): yield 1
    foo() | should.be.a('generator')

    class bar(object):
        def baz(self):
            pass

    bar | should.be.a('class')
    bar().baz | should.be.a('method')

    with pytest.raises(AssertionError):
        'foo' | should.be.an(int)

    with pytest.raises(AssertionError):
        [1, 2, 3] | should.be.a('tuple')



# -*- coding: utf-8 -*-
from ..operator import Operator


class TrueOperator(Operator):
    """
    Asserts if a given subject is `True` value.

    Example::

        # Should style
        True | should.be.true

        # Should style - negation form
        True | should.not_be.true

        # Expect style
        False | expect.to.be.true

        # Expect style - negation form
        False | expect.to_not.be.true
    """

    # Is the operator a keyword
    kind = Operator.Type.ACCESSOR

    # Operator keywords
    operators = ('true',)

    # Error message templates
    expected_message = Operator.Dsl.Message(
        'a value that is "True"',
        'a value that is not "True"',
    )

    # Subject message template
    subject_message = Operator.Dsl.Message(
        'a value of type "{type}" with content "{value}"',
    )

    def match(self, subject):
        if not self.ctx.negate and not isinstance(subject, bool):
            return False, ['subject is not a bool type']

        return subject is True, []


class FalseOperator(TrueOperator):
    """
    Asserts if a given subject is `False` value.

    Example::

        # Should style
        True | should.be.false

        # Should style - negation form
        True | should.not_be.false

        # Expect style
        False | expect.to.be.false

        # Expect style - negation form
        False | expect.to_not.be.false
    """

    # Is the operator a keyword
    kind = Operator.Type.ACCESSOR

    # Operator keywords
    operators = ('false',)

    # Error message templates
    expected_message = Operator.Dsl.Message(
        'a value that is "False"',
        'a value that is not "False"',
    )

    def match(self, subject):
        if not self.ctx.negate and not isinstance(subject, bool):
            return False, ['subject is not a bool type']

        return subject is False, []



# -*- coding: utf-8 -*-
import re

from ..operator import Operator


class MatchOperator(Operator):
    """
    Asserts if a given string matches a given regular expression.

    Example::

        # Should style
        'hello world' | should.match(r'Hello \w+')
        'hello world' | should.match(r'hello [A-Z]+', re.I))
        'hello world' | should.match.expression(r'hello [A-Z]+', re.I))

        # Should style - negation form
        'hello w0rld' | should.do_not.match(r'Hello \w+')
        'hello w0rld' | should.do_not.match(r'hello [A-Z]+', re.I))
        'hello world' | should.do_not.match.expression(r'hello [A-Z]+', re.I))

        # Expect style
        'hello world' | expect.to.match(r'Hello \w+')
        'hello world' | expect.to.match(r'hello [A-Z]+', re.I))
        'hello world' | expect.to.match.expression(r'hello [A-Z]+', re.I))

        # Expect style - negation form
        'hello w0rld' | expect.to_not.match(r'Hello \w+')
        'hello w0rld' | expect.to_not.match(r'hello [A-Z]+', re.I))
        'hello world' | expect.to_not.match.expression(r'hello [A-Z]+', re.I))
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('match', 'matches')

    # Operator chain aliases
    aliases = (
        'value', 'string', 'expression', 'token',
        'to', 'regex', 'regexp', 'word', 'phrase'
    )

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'a string that matches the expression "{value}"',
        'a string that does not match the expression "{value}"',
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with value "{value}"',
    )

    def match(self, subject, expected, *args):
        if not isinstance(subject, str):
            return False, ['subject must be an string, but got "{}"'.format(
                type(subject))]

        if not isinstance(expected, str):
            return False, [
                'value to match must be an string, but got "{}"'.format(
                    type(expected))
            ]

        return re.search(expected, subject, *args) is not None, []



# -*- coding: utf-8 -*-


class Context(object):
    """
    Context is used to keep state and share data across test operators.

    Context typically stores operator defined flags that are accessed later
    by other operators in order to retrieve information or modify its
    evaluation behavior.

    From test runner perspective, a new context is created on every
    assertion execution phase, so context data is context-dependent of
    the testing scope.

    Attributes:
        flags (dict): stores context shared data by flag name.

    Example::

        from grappa.context import Context

        ctx = Context()

        ctx.negate = True

        if ctx.has('negate'):
            ctx.negate = False
            ctx.value = 'foo'

        ctx.negate # => False
        ctx.value # => 'foo'
    """

    def __init__(self, defaults=None, **kw):
        self.__dict__['flags'] = defaults or {'negate': False}

        for name, value in kw.items():
            self.__dict__['flags'][name] = value

    def has(self, flag):
        """
        Returns `True` if the given flag name is present in the current
        context instance.

        Returns:
            bool
        """
        return flag in self.flags

    def __getattr__(self, name):
        """
        Overloads class attribute accessor method in order to proxy access to
        the store.
        """
        return self.flags.get(name)

    def __setattr__(self, name, value):
        """
        Overloads class set attribute method in order to proxy access to the
        store.
        """
        self.flags[name] = value

    def clone(self):
        """
        Returns a copy of the current `Context` instance.

        Returns:
            grappa.Context
        """
        return Context(self.__dict__['flags'].copy())

    def __repr__(self):
        """
        Human-readable context data.

        Returns:
            str
        """
        return str(self.__dict__['flags'])



# -*- coding: utf-8 -*-
from ..operator import Operator


class PresentOperator(Operator):
    """
    Asserts if a given subject is not ``None`` or a negative value
    if evaluated via logical unary operator.

    This operator is the opposite of ``empty``.

    Example::

        # Should style
        'foo' | should.be.present

        # Should style - negation form
        '' | should.not_be.present

        # Expect style
        'foo' | expect.to.be.present

        # Expect style - negation form
        False | expect.to_not.be.present
    """

    # Is the operator a keyword
    kind = Operator.Type.ACCESSOR

    # Operator keywords
    operators = ('present', 'exists')

    # Expected
    expected = 'a value which is not None and has data'

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'a value that is not None and its length is not "0"',
        'a value that is None or its length is "0"'
    )

    # Assertion information
    information = (
        Operator.Dsl.Help(
            Operator.Dsl.Description(
                'An present object is the oposite of an empty object.'
            )
        ),
    )

    def match(self, subject):
        if subject is None:
            return False, ['subject is "None"']

        return bool(subject is not None and subject), []



# -*- coding: utf-8 -*-
import collections
from ..operator import Operator


class ContainOperator(Operator):
    """
    Asserts if a given value or values can be found
    in a another object.

    Example::

        # Should style
        'foo bar' | should.contain('bar')
        ['foo', 'bar'] | should.contain('bar')
        ['foo', 'bar'] | should.contain('foo', 'bar')
        [{'foo': True}, 'bar'] | should.contain({'foo': True})

        # Should style - negation form
        'foo bar' | should.do_not.contain('bar')
        ['foo', 'bar'] | should.do_not.contain('baz')

        # Expect style
        'foo bar' | expect.to.contain('bar')
        ['foo', 'bar'] | expect.to.contain('bar')
        ['foo', 'bar'] | expect.to.contain('foo', 'bar')
        [{'foo': True}, 'bar'] | expect.to.contain({'foo': True})

        # Expect style - negation form
        'foo bar' | expect.to_not.contain('bar')
        ['foo', 'bar'] | expect.to_not.contain('baz')
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('contain', 'contains', 'includes')

    # Operator chain aliases
    aliases = ('value', 'item', 'string', 'text', 'expression', 'data')

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'a value that contains "{value}"',
        'a value that does not contains "{value}"',
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an value of type "{type}" with content "{value}"',
    )

    # Stores types to normalize before the assertion
    NORMALIZE_TYPES = (
        collections.Iterator,
        collections.MappingView,
        collections.Set
    )

    def match(self, subject, *expected):
        if isinstance(subject, self.NORMALIZE_TYPES):
            subject = list(subject)

        if self._is_not_a_sequence(subject):
            return False, ['is not a valid sequence type']

        return self._matches(subject, *expected)

    def _is_not_a_sequence(self, value):
        return not isinstance(value, collections.Sequence)

    def _matches(self, subject, *expected):
        reasons = []

        for expected_item in expected:
            matches_any, reason = self._matches_any(expected_item, subject)
            reasons.append(reason)

            if not matches_any:
                return False, [reason]

        return True, reasons

    def _matches_any(self, expected, subject):
        print(':::: subject ->', subject)
        if len(subject) == 0:
            return False, 'empty item'

        if isinstance(subject, str):
            if expected in subject:
                return True, 'item {0!r} found'.format(expected)
            return False, 'item {0!r} not found'.format(expected)

        for item in subject:
            print('::::', item, expected, item == expected)
            if item == expected:
                return True, 'item {0!r} found'.format(expected)

        return False, 'item {0!r} not found'.format(expected)



import pytest


def test_pass_function(should):
    def test(subject):
        return len(subject) == 3, []

    'foo' | should.pass_test(test)
    'fo' | should.do_not.pass_test(test)

    with pytest.raises(AssertionError):
        'foo' | should.do_not.pass_test(test)



# -*- coding: utf-8 -*-
import numbers
from ..operator import Operator


class EmptyOperator(Operator):
    """
    Asserts if a given subject is an empty object.

    A subject is considered empty if it's ``None``, ``0`` or ``len(subject)``
    is equals to ``0``.

    Example::

        # Should style
        [] | should.be.empty

        # Should style - negation form
        [1, 2, 3] | should.not_be.empty

        # Expect style
        tuple() | expect.to.be.empty

        # Expect style - negation form
        (1, 2, 3) | expect.to_not.be.empty
    """

    # Is the operator a keyword
    kind = Operator.Type.ACCESSOR

    # Operator keywords
    operators = ('empty',)

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'a value that is not "None" and its length is higher than zero'
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object with type "{type}" which its length cannot be measured'
    )

    # Assertion information
    information = (
        Operator.Dsl.Help(
            Operator.Dsl.Description(
                'An empty object can be "None", "0" or "len(x) == 0".',
                'Most objects in Python can be tested via "len(x)"',
                'such as str, list, tuple, dict, generator...',
                'as well as any object that implements "__len__()" method.',
            ),
            Operator.Dsl.Reference(
                'https://docs.python.org/3/library/functions.html#len'
            ),
        ),
    )

    def match(self, subject):
        if subject is None or subject is 0:
            return True

        if subject in (True, False):
            return False

        if isinstance(subject, numbers.Number) and subject != 0:
            return False

        try:
            return len(subject) == 0
        except TypeError:
            try:
                next(subject)
            except StopIteration:
                return True
            except Exception:
                return True
        return False



# -*- coding: utf-8 -*-
import inspect

from .log import log
from .empty import empty
from .base import BaseTest
from .engine import Engine
from .context import Context
from .resolver import OperatorResolver


class TestProxy(object):
    """
    TestProxy acts like a delegator intermediate proxy between public
    API calls and `Test` instances defining the test style.

    Arguments:
        style (str): test style between `should` or `expect`.
    """

    def __init__(self, style):
        self._style = style

    def __getattr__(self, name):
        """
        Overloads object attribute accessory proxy to the currently used
        operator and parent test instance.
        """
        _test = getattr(test, name)

        if inspect.ismethod(_test):
            return _test

        _test._ctx.style = self._style
        return _test

    def __call__(self, *args, **kw):
        """
        Makes proxy object itself callable, delegating the invokation
        to a new test instance.
        """
        _test = test(*args, **kw)
        _test._ctx.style = self._style
        return _test


class Test(BaseTest):
    """
    Test represents the test definition in `grappa` with extensible and
    dynamic, runtime inferred DSL based on registered operators and
    third-party plugins.

    Arguments:
        subject (mixed): subject value to test.
    """

    # Tracks context manager scopes
    _context = 0

    # Tracks yielded value by context manager
    _context_subject = empty

    # Global flag, only used by global singleton instance
    _global = False

    def __init__(self, subject=empty):
        self._engine = Engine()
        self._ctx = Context()
        self._ctx.subject = subject
        self._ctx.chained = False
        self._ctx.style = 'should'

    @property
    def should(self):
        """
        Alias name to self reference the current instance.
        Required for DSL API.
        """
        return self

    @property
    def expect(self):
        """
        Alias name to self reference the current instance.
        Required for DSL API.
        """
        return self

    @property
    def _root(self):
        return test

    def __call__(self, subject, overload=False):
        """
        Overloads function invokation of `Test` class instance.

        This is magical and widely used in `grappa` test execution by both
        developers and internal engine.

        Arguments:
            subject (mixed): test subject to use.
            overload (bool): `True` if the call if triggered via operator
                overloading invokation, otherise `False`.

        Returns:
            grappa.Test: new test instance with the given subject.
        """
        self._ctx.subject = subject
        return self._trigger() if overload else Test(subject)

    def __getattr__(self, name):
        """
        Overloads class attribute accessor proxying calls dynamically
        into assertion operators calls.

        This method is invoked by Python runtime engine, not by developers.
        """
        # Return a new test instance if running as global
        if self._global:
            # If using context manager, use context defined subject
            subject = self._context_subject if self._context else empty

            # Create new test and proxy attribute call
            return Test(subject).__getattr__(name)

        # Resolve and register operator by name
        return OperatorResolver(self).resolve(name)

    def _trigger(self):
        """
        Trigger assertions in the current test engine.

        Raises:
            AssertionError: in case of assertion error.
            Exception: in case of any other assertion error.
        """
        log.debug('[test] trigger with context: {}'.format(self._ctx))

        try:
            err = self._engine.run(self._ctx)
        except Exception as _err:
            err = _err
        finally:
            # Important: reset engine state to defaults
            self._engine.reset()
            self._root._engine.reset()

            # Update yielded value from asserion
            if self._ctx.value not in (None, empty):
                self._ctx.subject = self._ctx.value

        # If error is present, raise it!
        if err:
            raise err

        return self

    def _clone(self):
        """
        Clones the current `Test` instance.

        Returns:
            grappa.Test
        """
        test = Test(self._ctx.subject)
        test._ctx = self._ctx.clone()
        test._engine = self._engine.clone()
        return test

    def _flush(self):
        """
        Flushes the current test state, including test engine, assertions and
        current context.
        """
        self.__init__()

    # Assertions composition
    def all(self, *tests):
        """
        Composes multiple tests and executes them, in series, once a
        subject is received.

        Conditional composition operator equivalent to `all` built-in
        Python function.

        Arguments:
            *tests (grappa.Test): test instances to run.
        """
        def run_tests(subject):
            for test in tests:
                try:
                    test(subject, overload=True)
                except Exception as err:
                    return err
            return True
        self._engine.add_assertion(run_tests)
        return self

    def any(self, *tests):
        """
        Composes multiple tests and executes them, in series, once a
        subject is received.

        Conditional composition operator equivalent to `any` built-in
        Python function.

        Arguments:
            *tests (grappa.Test): test instances to run.
        """
        def run_tests(subject):
            err = None
            for test in tests:
                try:
                    test(subject, overload=True)
                except Exception as _err:
                    err = _err
                else:
                    return True
            return err

        self._engine.add_assertion(run_tests)
        return self

    def __overload__(self, subject):
        """
        Method triggered by magic methods executed via operator overloading.
        """
        if isinstance(subject, Test):
            # Clone test instance to make it side-effects free
            fork = subject._clone()
            fork._ctx.chained = True
            fork._ctx.subject = self._ctx.subject
            # Trigger assertions
            return fork._trigger()

        # Otherwise invoke the test function with a subject
        return self.__call__(subject, overload=True)

    def __or__(self, value):
        """
        Overloads ``|`` as from left-to-right operator precedence expression.
        """
        return self.__overload__(value)

    def __ror__(self, value):
        """
        Overloads ``|`` operator.
        """
        return self.__overload__(value)

    def __gt__(self, value):
        """
        Overloads ``>`` operator.
        """
        return self.__overload__(value)

    def __enter__(self):
        """
        Initializes context manager.
        """
        log.debug('creates new test context manager: {}'.format(self._ctx))
        test._context += 1
        test._context_subject = self._ctx.subject

    def __exit__(self, etype, value, traceback):
        """
        Exists context manager.
        """
        log.debug('exists test context manager: {}'.format(value))
        test._context -= 1

        if test._context == 0:
            test._context_subject = empty


# Create global singleton instance
test = Test()

# This is black magic in order to deal with chainable states
# and operator precedence.
test._global = True



import pytest
from grappa.operators.none import NoneOperator


def test_should_none(should):
    None | should.be.none
    'foo' | should.not_be.none

    with pytest.raises(AssertionError):
        False | should.be.none

    with pytest.raises(AssertionError):
        None | should.not_be.none

    with pytest.raises(AssertionError):
        'foo' | should.be.none

    with pytest.raises(AssertionError):
        [1, 2, 3] | should.be.none


def test_expect_none(expect):
    None | expect.to.be.none

    with pytest.raises(AssertionError):
        False | expect.to.be.none

    with pytest.raises(AssertionError):
        'foo' | expect.to.be.none

    with pytest.raises(AssertionError):
        [1, 2, 3] | expect.to.be.none


def test_none_operator(ctx):
    assert NoneOperator(ctx).match(None) == (True, [])
    assert NoneOperator(ctx).match(1) == (False, [])



# -*- coding: utf-8 -*-
import inspect
from ..operator import Operator


class PassTestOperator(Operator):
    """
    Asserts if a given function passes the test.

    Example::

        # Should style
        'foo' | should.pass_test(lambda x: len(x) > 2)
        [1, 2, 3] | should.pass_function(lambda x: 2 in x)

        # Should style - negation form
        'foo' | should.do_not.pass_test(lambda x: len(x) > 3)
        [1, 2, 3] | should.do_not.pass_function(lambda x: 5 in x)

        # Expect style
        'foo' | expect.to.pass_test(lambda x: len(x) > 2)
        [1, 2, 3] | expect.to.pass_function(lambda x: 2 in x)

        # Expect style - negation form
        'foo' | expect.to_not.pass_test(lambda x: len(x) > 3)
        [1, 2, 3] | expect.to_not.pass_function(lambda x: 5 in x)
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('pass_test', 'pass_function')

    # Expected message template
    expected_message = Operator.Dsl.Message(
        'a value that passes the test function',
        'a value that does not pass the test function',
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an value of type "{type}" with value "{value}"',
    )

    def match(self, subject, fn):
        if not any([inspect.isfunction(fn) or inspect.ismethod(fn)]):
            return False, ['matcher argument must be a function or method']
        return fn(subject)



# -*- coding: utf-8 -*-
from ..operator import Operator


class WithinOperator(Operator):
    """
    Asserts that a number is within a range.

    Example::

        # Should style
        4 | should.be.within(2, 5)
        5 | should.be.between(2, 5)
        4.5 | should.be.within(4, 5)

        # Should style - negation form
        4 | should.not_be.within(2, 5)
        5 | should.not_be.between(2, 5)
        4.5 | should.not_be.within(4, 5)

        # Expect style
        4 | expect.to.be.within(2, 5)
        5 | expect.to.be.between(2, 5)
        4.5 | expect.to.be.within(4, 5)

        # Expect style - negation form
        4 | expect.to_not.be.within(2, 5)
        5 | expect.to_not.be.between(2, 5)
        4.5 | expect.to_not.be.within(4, 5)
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('within', 'between')

    # Operator keywords
    aliases = ('to', 'numbers', 'range')

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'a number that is within range {value}',
        'a number that is not within range {value}'
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with value "{value}"',
    )

    def match(self, subject, start, end):
        if self.ctx.length:
            subject = len(subject)

        if not isinstance(subject, (int, float, complex)):
            return False, ['subject is not a numeric type']

        return start <= subject <= end, []



# -*- coding: utf-8 -*-
from .reporter import ErrorReporter


class Runner(object):
    """
    Runner is responsible of triggering the registed assertion operators in the
    current engine.

    Arguments:
        engine (grappa.Engine)
    """

    def __init__(self, engine):
        self.engine = engine

    def render_error(self, ctx, error):
        # Expose keywords via context (this should be improved)
        ctx.keywords = self.engine.keywords
        # Render error exception
        return ErrorReporter(ctx).run(error)

    def run_assertions(self, ctx):
        # Trigger assertion functions
        for assertion in self.engine.assertions:
            # Run assertion with the given subject
            result = assertion(ctx.subject)

            # If assertion passed, just continue with it
            if result is True:
                continue

            # Forward original grappa error
            if all([isinstance(result, AssertionError),
                    hasattr(result, '__grappa__')]):
                return result

            # Otherwise render assertion error accordingly
            return self.render_error(ctx, result)

    def run(self, ctx):
        """
        Runs the current phase.
        """
        # Reverse engine assertion if needed
        if ctx.reverse:
            self.engine.reverse()

        if self.engine.empty:
            raise AssertionError('grappa: no assertions to run')

        try:
            # Run assertion in series and return error, if present
            return self.run_assertions(ctx)
        except Exception as _err:
            # Handle legit grappa internval errors
            if getattr(_err, '__legit__', False):
                raise _err
            # Otherwise render it
            return self.render_error(ctx, _err)



import pytest


def test_should_contain(should):
    'hello world' | should.contain('world') | should.contain('hello')
    'hello world' | should.contain('w') | should.contain('o')

    [1, 2, 3] | should.contain(1) | should.contain(3)
    ('foo', 'bar', 123) | should.contain('bar') | should.contain(123)
    [{'foo': 1}] | should.contain({'foo': 1})

    with pytest.raises(AssertionError):
        'hello world' | should.contain('planet')

    with pytest.raises(AssertionError):
        [1, 2, 3] | should.contain(4)



# -*- coding: utf-8 -*-
import logging
import functools

from .log import log


def validate(method):
    """
    Config option name value validator decorator.
    """
    # Name error template
    name_error = 'configuration option "{}" is not supported'

    @functools.wraps(method)
    def validator(self, name, *args):
        if name not in self.allowed_opts:
            raise ValueError(name_error.format(name))
        return method(self, name, *args)
    return validator


class Config(object):
    """
    Config provides a simple key-value Pythonic data store for grappa
    runtime configuration settings.
    """

    # Default options store
    defaults = {
        # Debug enables internal logging. Defaults to False.
        'debug': False,
        # show_code enables/disables code output in errors.
        'show_code': True,
        # colors enables/disables colored ANSI sequences for the terminal
        'use_colors': True
    }

    # List of supported configuration names
    allowed_opts = defaults.keys()

    def __init__(self):
        self.__dict__['opts'] = self.defaults.copy()

    @validate
    def __getattr__(self, name):
        return self.opts.get(name)

    @validate
    def __setattr__(self, name, value):
        self.opts[name] = value

        if name == 'debug':
            log.setLevel(logging.DEBUG if value is True
                         else logging.CRITICAL)

        log.debug('[config] set option {}="{}"'.format(name, value))


# Global config store
config = Config()



# -*- coding: utf-8 -*-
from ..operator import Operator


class LengthOperator(Operator):
    """
    Asserts that a given iterable object has exact length.

    Example::

        # Should style
        'foo' | should.have.length(3)
        [1, 2, 3] | should.have.length.of(3)
        iter([1, 2, 3]) | should.have.length.equal.to(3)

        # Should style - negation form
        'foobar' | should.not_have.length(3)
        [1, 2, 3, 4] | should.not_have.length.of(3)
        iter([1, 2, 3, 4]) | should.not_have.length.equal.to(3)

        # Expect style
        'foo' | expect.to.have.length(3)
        [1, 2, 3] | expect.to.have.length.of(3)
        iter([1, 2, 3]) | expect.to.have.length.equal.to(3)

        # Expect style - negation form
        'foobar' | expect.to_not.have.length(3)
        [1, 2, 3, 4] | expect.to_not.have.length.of(3)
        iter([1, 2, 3, 4]) | expect.to_not.have.length.equal.to(3)
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('length', 'size')

    # Operator chain aliases
    aliases = ('equal', 'to', 'of')

    # Error message templates
    expected_message = Operator.Dsl.Message(
        'an object that its length is equal to {value}',
        'an object that its length is not equal to {value}',
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with length {length}'
    )

    information = (
        Operator.Dsl.Help(
            Operator.Dsl.Description(
                'Object length is measured by using "len()" built-in',
                'Python function or consuming an lazy iterable, such as a'
                'generator. Most built-in types and objects in Python',
                'can be tested that way, such as str, list, tuple, dict...',
                'as well as any object that implements "__len__()" method.'
            ),
            Operator.Dsl.Reference(
                'https://docs.python.org/3/library/functions.html#len'
            ),
        ),
    )

    def __init__(self, *args, **kw):
        Operator.__init__(self, *args, **kw)
        self.ctx.length = True

    def match(self, subject, expected):
        try:
            return len(subject) == expected, [
                'unexpected object length: {}'.format(len(subject))]
        except TypeError:
            try:
                return len([i for i in subject]) == expected
            except Exception:
                pass
        return False, ['cannot measure the length of the given object']



# -*- coding: utf-8 -*-
import inspect

from .log import log
from .engine import Engine


def use(plugin):
    """
    Register plugin in grappa.

    `plugin` argument can be a function or a object that implement `register`
    method, which should accept one argument: `grappa.Engine` instance.

    Arguments:
        plugin (function|module): grappa plugin object to register.

    Raises:
        ValueError: if `plugin` is not a valid interface.

    Example::

        import grappa

        class MyOperator(grappa.Operator):
            pass

        def my_plugin(engine):
            engine.register(MyOperator)

        grappa.use(my_plugin)
    """
    log.debug('register new plugin: {}'.format(plugin))

    if inspect.isfunction(plugin):
        return plugin(Engine)

    if plugin and hasattr(plugin, 'register'):
        return plugin.register(Engine)

    raise ValueError('invalid plugin: must be a function or '
                     'implement register() method')



import pytest
from grappa.operators.empty import EmptyOperator


def test_should_empty(should):
    None | should.be.empty
    [] | should.be.empty
    '' | should.be.empty
    tuple() | should.be.empty
    0 | should.be.empty
    iter([]) | should.be.empty

    with pytest.raises(AssertionError):
        False | should.be.empty

    with pytest.raises(AssertionError):
        True | should.be.empty

    with pytest.raises(AssertionError):
        'foo' | should.be.empty

    with pytest.raises(AssertionError):
        [1, 2, 3] | should.be.empty

    with pytest.raises(AssertionError):
        iter([1, 2, 3]) | should.be.empty


def test_expect_empty(expect):
    None | expect.to.be.empty
    [] | expect.to.be.empty
    '' | expect.to.be.empty
    tuple() | expect.to.be.empty
    0 | expect.to.be.empty
    iter([]) | expect.to.be.empty

    with pytest.raises(AssertionError):
        False | expect.to.be.empty

    with pytest.raises(AssertionError):
        True | expect.to.be.empty

    with pytest.raises(AssertionError):
        'foo' | expect.to.be.empty

    with pytest.raises(AssertionError):
        [1, 2, 3] | expect.to.be.empty

    with pytest.raises(AssertionError):
        iter([1, 2, 3]) | expect.to.be.empty


def test_empty_operator(ctx):
    assert EmptyOperator(ctx).match(0) is True
    assert EmptyOperator(ctx).match(None) is True
    assert EmptyOperator(ctx).match('') is True
    assert EmptyOperator(ctx).match(iter([])) is True

    assert EmptyOperator(ctx).match(True) is False
    assert EmptyOperator(ctx).match(False) is False
    assert EmptyOperator(ctx).match('foo') is False
    assert EmptyOperator(ctx).match(123) is False
    assert EmptyOperator(ctx).match(0.2321) is False
    assert EmptyOperator(ctx).match(iter([1, 2, 3])) is False


def test_empty_operator_properties(should):
    (EmptyOperator
        | should.have.property('kind')
        > should.be.equal.to('accessor'))

    (EmptyOperator
        | should.have.property('operators')
        > should.have.length.of(1)
        > should.be.equal.to(('empty',)))

    EmptyOperator | should.have.property('aliases') > should.be.empty

    EmptyOperator | should.have.property('expected_message')
    EmptyOperator | should.have.property('subject_message')

    (EmptyOperator | should.have.property('information')
        > should.be.a('tuple')
        > should.have.length.of(1))



import pytest


def test_expect_below(should):
    10 | should.be.below(100)
    1 | should.be.below.of(2)
    10 | should.be.lower.than(100)

    # negation
    10 | should.not_be.below.of(2)
    10 | should.not_be.lower.than(2)

    with pytest.raises(AssertionError):
        10 | should.be.below.of(2)

    with pytest.raises(AssertionError):
        None | should.be.below.of(10)


def test_expect_above(should):
    10 | should.be.above(9)
    3 | should.be.above.of(2)

    # negation
    10 | should.not_be.above.of(11)

    with pytest.raises(AssertionError):
        10 | should.be.above.of(100)

    with pytest.raises(AssertionError):
        None | should.be.above.of(10)


def test_expect_below_or_equal(should):
    10 | should.be.below_or_equal(10)
    5 | should.be.below_or_equal(10)

    # negation
    10 | should.not_be.below_or_equal.to(5)

    with pytest.raises(AssertionError):
        10 | should.not_be.below_or_equal.to(10)

    with pytest.raises(AssertionError):
        None | should.be.below_or_equal.of(10)


def test_expect_above_or_equal(should):
    10 | should.be.above_or_equal(10)
    101 | should.be.above_or_equal(100)
    3 | should.be.above_or_equal.to(3)

    # negation
    10 | should.not_be.above_or_equal.to(100)

    with pytest.raises(AssertionError):
        10 | should.not_be.above_or_equal.to(10)

    with pytest.raises(AssertionError):
        None | should.be.above_or_equal.of(10)



# -*- coding: utf-8 -*-
import inspect
import functools
from .engine import Engine
from .operator import Operator


# Explicit module members to export
__all__ = (
    'operator',
    'attribute',
    'accessor',
    'register'
)


def operator(name=None, operators=None, aliases=None, kind=None):
    """
    Registers a new operator function in the test engine.

    Arguments:
        *args: variadic arguments.
        **kw: variadic keyword arguments.

    Returns:
        function
    """
    def delegator(assertion, subject, expected, *args, **kw):
        return assertion.test(subject, expected, *args, **kw)

    def decorator(fn):
        operator = Operator(fn=fn, aliases=aliases, kind=kind)
        _name = name if isinstance(name, str) else fn.__name__
        operator.operators = (_name,)

        _operators = operators
        if isinstance(_operators, list):
            _operators = tuple(_operators)

        if isinstance(_operators, tuple):
            operator.operators += _operators

        # Register operator
        Engine.register(operator)
        return functools.partial(delegator, operator)

    return decorator(name) if inspect.isfunction(name) else decorator


def attribute(*args, **kw):
    """
    Registers a new attribute only operator function in the test engine.

    Arguments:
        *args: variadic arguments.
        **kw: variadic keyword arguments.

    Returns:
        function
    """
    return operator(kind=Operator.Type.ATTRIBUTE, *args, **kw)


def accessor(*args, **kw):
    """
    Registers a new keyword only operator function in the test engine.

    Arguments:
        *args: variadic arguments.
        **kw: variadic keyword arguments.

    Returns:
        function
    """
    return operator(kind=Operator.Type.ACCESSOR, *args, **kw)


def register(operator):
    """
    Registers a new operator class in the test engine.

    Arguments:
        operator (Operator): operator class to register.

    Returns:
        Operator: operator class.
    """
    Engine.register(operator)
    return operator



import pytest


def test_should_within(should):
    10 | should.be.between(5, 10)
    101 | should.be.within(100, 102)

    # negation
    11 | should.not_be.between(5, 9)

    with pytest.raises(AssertionError):
        10 | should.not_be.between.to(5, 10)

    with pytest.raises(AssertionError):
        None | should.be.between.numbers(10, 10)



# -*- coding: utf-8 -*-
from ..engine import Engine

# Module symbols to export
__all__ = ('operators', 'load')


# List of built-in operators
operators = (
    # Module name   # Operator class to import
    ('attributes', ),
    ('type',        'TypeOperator'),
    ('none',        'NoneOperator'),
    ('keys',        'KeysOperator'),
    ('match',       'MatchOperator'),
    ('length',      'LengthOperator'),
    ('empty',       'EmptyOperator'),
    ('equal',       'EqualOperator'),
    ('within',      'WithinOperator'),
    ('present',     'PresentOperator'),
    ('contain',     'ContainOperator'),
    ('callable',    'CallableOperator'),
    ('property',    'PropertyOperator'),
    ('pass_test',   'PassTestOperator'),
    ('implements',  'ImplementsOperator'),
    ('raises',      'RaisesOperator'),

    ('bool',        'TrueOperator', 'FalseOperator'),
    ('start_end',   'StartWithOperator', 'EndWithOperator'),

    ('range',       'BelowOperator', 'AboveOperator',
                    'AboveOrEqualOperator', 'BelowOrEqualOperator'),
)


def load():
    """
    Loads the built-in operators into the global test engine.
    """
    for operator in operators:
        module, symbols = operator[0], operator[1:]
        path = 'grappa.operators.{}'.format(module)

        # Dynamically import modules
        operator = __import__(path, None, None, symbols)

        # Register operators in the test engine
        for symbol in symbols:
            Engine.register(getattr(operator, symbol))



# -*- coding: utf-8 -*-

# Import test module
from .test import Test, TestProxy
# Expose plugin module as part of the public API
from .plugin import use
# Expose config module as part of the public API
from .config import config
# Expose decorator for operators factory
from .decorators import *  # noqa
# Expose Operatocdr base class
from .operator import Operator
# Required to load built-in operators
from . import operators

# Initialize colorama stdout/stderr interceptor
import colorama
colorama.init()

# IMPORTANT! Autoload built-in operators
operators.load()

# Global test instance for "should" testing declaration style
should = TestProxy('should')

# Global test instance for "expect" testing declaration style
expect = TestProxy('expect')

# Module symbols to export as public API
__all__ = (
    'should',
    'expect',
    'Test',
    'use',
    'config',
    'Operator',
    'attribute',
    'operator',
    'accessor',
    'register'
)



# -*- coding: utf-8 -*-
from ..operator import Operator


class CallableOperator(Operator):
    """
    Asserts if a given subject is a callable type or an object that
    implements ``__call__()`` magic method.

    Example::

        # Should style
        (lambda x: x) | should.be.callable

        # Should style - negation form
        None | should.not_be.callable

        # Expect style
        (lambda x: x) | expect.to.be.callable

        # Expect style - negation form
        None | expect.to_not.be.callable
    """

    # Is the operator a keyword
    kind = Operator.Type.ACCESSOR

    # Operator keywords
    operators = ('callable',)

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'a callable object, such as a function or method',
        'a non callable object, such as a function or method'
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object which is a "{type}" type'
    )

    # Assertion information
    information = (
        Operator.Dsl.Help(
            Operator.Dsl.Description(
                'Callable objects can be tested via "callable(x)" Python',
                'built-in function.'
            ),
            Operator.Dsl.Reference(
                'https://docs.python.org/3/library/functions.html#callable'
            ),
        ),
        Operator.Dsl.Help(
            Operator.Dsl.Description(
                'Callable objects in Python are functions, methods, classes,',
                'or any other object that implements "__call__()" method.'
            ),
            Operator.Dsl.Reference(
                'http://stackoverflow.com/a/2436614'
            ),
        ),
    )

    def match(self, value):
        if value is None:
            return False, ['a callable value cannot be "None"']

        # Custom expectations messages
        self.expected = 'an callable object, such as a function or method'
        self.value = value

        return callable(value), []



# -*- coding: utf-8 -*-
from .empty import empty
from .assertion import AssertionProxy


class OperatorResolver(object):
    """
    Resolves and triggers an operator based on its name.

    This class is highly-coupled to `grappa.Test` and consumes `grappa.Engine`
    and `grappa.Context` in order to trigger operator resolution logic.
    """

    def __init__(self, test):
        self.test = test
        self.ctx = test._ctx
        self.engine = test._engine

    def run_attribute(self, operator):
        operator.run()

    def run_accessor(self, operator):
        # Register assertion function
        def assertion(subject):
            self.ctx.subject = subject
            return operator.run(subject)

        # Add assertion function
        self.engine.add_assertion(assertion)

        # Self-trigger tests if running as global
        if self.ctx.chained:
            self.test._trigger()

        return self.test

    def run_matcher(self, operator):
        # Process assert operators
        def wrapper(*expected, **kw):
            # Retrieve optional custom assertion message
            if 'msg' in kw:
                # Set user-defined message
                self.ctx.message = kw.pop('msg')

            def assertion(subject):
                operator.ctx.subject = subject
                operator.ctx.expected = expected
                return operator.run(subject, *expected, **kw)

            # Register assertion function
            self.test._engine.add_assertion(assertion)

            # Trigger tests on function call if running as chained call
            if self.ctx.chained or self.ctx.subject is not empty:
                return self.test._trigger()

            return self.test

        return AssertionProxy(self.test, operator, wrapper)

    def resolve(self, name):
        # Find an assertion operator by name
        operator = self.engine.find_operator(name)
        if not operator:
            raise AttributeError(
                '"{}" has no assertion operator called "{}"'.format(
                    self.ctx.style, name
                ))

        # Register attribute access
        self.engine.add_keyword(name)

        # Create operator instance with current context
        operator = operator(context=self.ctx, operator_name=name)

        # Reset context sequence
        if self.ctx.reset:
            self.engine.reset_keywords()
            self.ctx.reset = False
            self.ctx.reverse = True

        # Dynamically retrieve operator
        method_name = 'run_{}'.format(operator.kind)
        run_operator = getattr(self, method_name, None)

        # If operator kind is not support, raise an exception
        if not run_operator:
            raise ValueError('operator "{}" has not a valid kind "{}"'.format(
                operator.__class__.__name__,
                operator.kind
            ))

        # Register operator assertion for lazy execution
        return run_operator(operator) or self.test



# -*- coding: utf-8 -*-
import inspect
from ..operator import Operator


class RaisesOperator(Operator):
    """
    Asserts if a given function raises an exception.

    Example::

        def fn():
            raise ValueError('error')

        # Should style
        fn | should.raise_error()
        fn | should.raise_error(ValueError)
        fn | should.raise_error(AttributeError, ValueError)

        # Should style - negation form
        fn | should.do_not.raise_error()
        fn | should.do_not.raise_error(ValueError)
        fn | should.do_not.raise_error(AttributeError, ValueError)

        # Expect style
        fn | expect.to.raise_error()
        fn | expect.to.raise_error(ValueError)
        fn | expect.to.raise_error(AttributeError, ValueError)

        # Expect style - negation form
        fn | expect.to_not.raise_error()
        fn | expect.to_not.raise_error(ValueError)
        fn | expect.to_not.raise_error(AttributeError, ValueError)
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('raises', 'raise_error', 'raise_errors')

    # Operator chain aliases
    aliases = ('to', 'that', 'are', 'instance', 'of')

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'a callable object that raises the exception(s) "{value}"',
        'a callable object that do not raise the exception(s) "{value}"',
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with reference "{value}"',
    )

    def match(self, fn, *errors):
        if not any([inspect.isfunction(fn) or inspect.ismethod(fn)]):
            return False, ['subject must be a function or method']

        try:
            fn()
        except Exception as err:
            self.value = err

            # If no errors present, just raise the exception
            if not errors:
                return True, []

            # Otherwise match errors
            return isinstance(err, *errors), ['invalid raised exception']
        else:
            return False, ['did not raise any exception']



import pytest
from grappa.operators.callable import CallableOperator


def test_should_callable(should):
    test_should_callable | should.be.callable
    (lambda x: x) | should.be.callable
    CallableOperator | should.be.callable
    CallableOperator.match | should.be.callable

    with pytest.raises(AssertionError):
        tuple() | should.be.callable

    with pytest.raises(AssertionError):
        0 | should.be.callable


def test_expect_callable(expect):
    test_expect_callable | expect.to.be.callable
    (lambda x: x) | expect.to.be.callable
    CallableOperator | expect.to.be.callable
    CallableOperator.match | expect.to.be.callable

    with pytest.raises(AssertionError):
        tuple() | expect.to.be.callable

    with pytest.raises(AssertionError):
        0 | expect.to.be.callable


def test_callable_operator(ctx):
    assert CallableOperator(ctx).match(lambda x: x) == (True, [])
    assert CallableOperator(ctx).match(CallableOperator) == (True, [])
    assert CallableOperator(ctx).match(CallableOperator.match) == (True, [])

    assert CallableOperator(ctx).match(0) == (False, [])
    assert CallableOperator(ctx).match('foo') == (False, [])
    assert CallableOperator(ctx).match(iter([1, 2, 3])) == (False, [])
    assert CallableOperator(ctx).match(None) == (
        False, ['a callable value cannot be "None"'])


def test_callable_operator_properties(should):
    (CallableOperator
        | should.have.property('kind')
        > should.be.equal.to('accessor'))

    (CallableOperator
        | should.have.property('operators')
        > should.have.length.of(1)
        > should.be.equal.to(('callable',)))

    CallableOperator | should.have.property('aliases') > should.be.empty

    CallableOperator | should.have.property('expected_message')
    CallableOperator | should.have.property('subject_message')

    (CallableOperator | should.have.property('information')
        > should.be.a('tuple')
        > should.have.length.of(2))



# -*- coding: utf-8 -*-
import inspect
import functools

from .log import log
from .empty import empty
from . import operator_dsl as OperatorDsl
from .exceptions import GrappaAssertionError


class OperatorTypes(object):
    """
    OperatorTypes is used as struct to store the operator types flags.
    """
    # Attribute operators have no validation logic and only defines chained
    # API DSL keywords, typically for test declaration purposes.
    ATTRIBUTE = 'attribute'

    # Accessor operators do not accept expectation arguments and only does
    # static validation logic. E.g: `true/false` operators.
    ACCESSOR = 'accessor'

    # Matcher operators accepts expectation arguments and does the validation
    # based on them. E.g: `equal`.
    MATCHER = 'matcher'


class Operator(object):
    """
    Operator implements a base class with common logic and required interface
    that is used by specific operator implementations.

    Any operator should inherit from this class.

    Attributes:
        Dsl (grappa.operators_dsl): DSL for operator error messages templating
        Type (grappa.operators.OperatorTypes): support operators types
        kind (str): stores the operator kind
        operators (list|tuple): operator keywords
        aliases (list|tuple): chain attributes aliases for expressivity
        information (list|tuple): optional additional help in case of error
        subject_message (str|Operator.Dsl.Message): optional subject message
        expected_message (str|Operator.Dsl.Message): optional expected message
        operator_name (str): invokation operator name keyword
        suboperators (list|tuple[grappa.Operator]): optional child operators
    """

    # Shortcut alias for operator DSL module
    Dsl = OperatorDsl

    # Shortcut to the supported operator types
    Type = OperatorTypes

    # Stores operator kind
    kind = OperatorTypes.MATCHER

    # Operator keywords
    operators = tuple()

    # Chain alias keywords
    aliases = tuple()

    # Stores assertion error tips
    information = []

    # Stores test execution context
    ctx = None

    # Stores the selected operator name when executed
    operator_name = None

    # Stores optional suboperators (not supported yet)
    suboperators = []

    # Stores optionally yielded operator result value
    value = empty

    # Stores operator assertion expectation
    expected = empty

    # Default subject error template
    subject_message = Dsl.Message(
        'a value of "{type}" type with content "{value}"'
        'a value of "{type}" type with content "{value}"'
    )

    # Default expected error template
    expected_message = Dsl.Message(
        'a value of "{type}" type that satisfies assertion with "{value}"',
        'a value of "{type}" type that does not satisfies assertion '
        'with "{value}"',
    )

    def __init__(self, context=None, operator_name=None, fn=None,
                 kind=None, aliases=None, operators=None, suboperators=None):
        if inspect.isfunction(fn):
            self.match = fn
        if kind:
            self.kind = kind
        if operators:
            self.operators = operators
        if aliases:
            self.aliases = aliases
        if context:
            self.ctx = context
        if operator_name:
            self.operator_name = operator_name
        if suboperators:
            self.suboperators = suboperators

    def match(self, *args, **kw):
        raise NotImplementedError('"match" method must be implemented')

    def __call__(self, *args, **kw):
        """
        Overloads function invokation in current operator and creates a new
        cloned operator instance.
        """
        operator = Operator()
        operator.__dict__ = self.__dict__.copy()
        operator.__init__(*args, **kw)
        return operator

    def observe(matcher):
        """
        Internal decorator to trigger operator hooks before/after
        matcher execution.
        """
        @functools.wraps(matcher)
        def observer(self, subject, *expected, **kw):
            # Trigger before hook, if present
            if hasattr(self, 'before'):
                self.before(subject, *expected, **kw)

            # Trigger matcher method
            result = matcher(self, subject, *expected, **kw)

            # After error hook
            if result is not True and hasattr(self, 'after_error'):
                self.after_error(result, subject, *expected, **kw)

            # After success hook
            if result is True and hasattr(self, 'after_success'):
                self.after_success(subject, *expected, **kw)

            return result
        return observer

    def _make_error(self, reasons=None, error=None):
        return GrappaAssertionError(
            error=error,
            reasons=reasons,
            operator=self
        )

    @observe
    def run_matcher(self, subject, *expected, **kw):
        """
        Runs the operator matcher test function.
        """
        # Update assertion expectation
        self.expected = expected

        _args = (subject,)
        if self.kind == OperatorTypes.MATCHER:
            _args += expected

        try:
            result = self.match(*_args, **kw)
        except Exception as error:
            return self._make_error(error=error)

        reasons = []
        if isinstance(result, tuple):
            result, reasons = result

        if result is False and self.ctx.negate:
            return True

        if result is True and not self.ctx.negate:
            return True

        return self._make_error(reasons=reasons)

    def run(self, *args, **kw):
        """
        Runs the current operator with the subject arguments to test.

        This method is implemented by matchers only.
        """
        log.debug('[operator] run "{}" with arguments: {}'.format(
            self.__class__.__name__, args
        ))

        if self.kind == OperatorTypes.ATTRIBUTE:
            return self.match(self.ctx)
        else:
            return self.run_matcher(*args, **kw)

    def __enter__(self):
        raise NotImplementedError('operator cannot be used like "with" statement')  # noqa

    def __exit__(self, etype, value, traceback):
        self.__enter__()



import pytest


def test_expect_length(should):
    [] | should.have.length.of(0)
    'foo' | should.have.length.of(3) | should.have.length(3)
    'foo' | should.not_have.length.of(4)
    'foo' | should.not_have.length.of(0) | should.not_have.length.of(2)

    with pytest.raises(AssertionError):
        'foo' | should.have.length(0)

    with pytest.raises(AssertionError):
        'foo' | should.have.length.of(3) | should.have.length(5)

    with pytest.raises(AssertionError):
        None | should.have.length(0)

    with pytest.raises(AssertionError):
        'foo' | should.not_have.length.of(3)



# -*- coding: utf-8 -*-


class Empty(object):
    """
    Empty object represents emptyness state in `grappa`.
    """

    def __repr__(self):
        return 'Empty'

    def __len__(self):
        return 0


# Object reference representing emptpyness
empty = Empty()



# -*- coding: utf-8 -*-
from .runner import Runner


def isoperator(x):
    """
    Returns `True` if the given object implements the required attributes for
    an operator.

    Returns:
        bool
    """
    return all(
        hasattr(x, name) for name in ('run', 'operators', 'kind', '__class__')
    )


def register_operators(*operators):
    """
    Registers one or multiple operators in the test engine.
    """
    def validate(operator):
        if isoperator(operator):
            return True

        raise NotImplementedError('invalid operator: {}'.format(operator))

    def register(operator):
        # Register operator by DSL keywords
        for name in operator.operators:
            # Check valid operators
            if name in Engine.operators:
                raise ValueError('operator name "{}" from {} is already '
                                 'in use by other operator'.format(
                                    name,
                                    operator.__name__
                                 ))

            # Register operator by name
            Engine.operators[name] = operator

    # Validates and registers operators
    [register(operator) for operator in operators if validate(operator)]


class Engine(object):
    """
    Engine implements the test engine responsible of registering, storing and
    running test operators.
    """

    # globally store test operators by name
    operators = {}

    def __init__(self):
        self.keywords = []
        self.assertions = []

    @staticmethod
    def register(*operators):
        """
        Registers one or multiple operators in the test engine.
        """
        register_operators(*operators)

    @property
    def empty(self):
        return len(self.assertions) == 0

    def add_assertion(self, assertion):
        self.assertions.append(assertion)

    def add_keyword(self, keyword):
        self.keywords.append(keyword)

    def reverse(self):
        self.assertions.reverse()

    def find_operator(self, name):
        return self.operators.get(name)

    def run(self, ctx):
        return Runner(self).run(ctx)

    def reset(self):
        self.__init__()

    def reset_keywords(self):
        self.keywords = []

    def clone(self):
        engine = Engine()
        engine.keywords = self.keywords[:]
        engine.assertions = self.assertions[:]
        return engine



# -*- coding: utf-8 -*-
import collections

from ..operator import Operator
from ..constants import STR_TYPES


class StarEndBaseOpeartor(Operator):

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Chain aliases
    aliases = (
        'word', 'string', 'number', 'name', 'numbers', 'items',
        'value', 'char', 'letter', 'by', 'character', 'item',
    )

    def match(self, subject, *expected):
        if self.is_unordered_dict(subject):
            return False, ['does not have ordered keys']

        return self.matches(subject, *expected)

    def is_unordered_dict(self, subject):
        if isinstance(subject, collections.Mapping):
            if not hasattr(collections, 'OrderedDict'):
                return True
            return not isinstance(subject, collections.OrderedDict)
        return False


class StartWithOperator(StarEndBaseOpeartor):
    """
    Asserts if a given value starts with a specific items.

    Example::

        # Should style
        'foo' | should.start_with('f')
        'foo' | should.start_with('fo')
        [1, 2, 3] | should.start_with.number(1)
        iter([1, 2, 3]) | should.start_with.numbers(1, 2)
        OrderedDict([('foo', 0), ('bar', 1)]) | should.start_with.item('foo')

        # Should style - negation form
        'foo' | should.do_not.start_with('o')
        'foo' | should.do_not.start_with('o')
        [1, 2, 3] | should.do_not.start_with(2)
        iter([1, 2, 3]) | should.do_not.start_with.numbers(3, 4)
        OrderedDict([('foo', 0), ('bar', 1)]) | should.start_with('bar')

        # Expect style
        'foo' | expect.to.start_with('f')
        'foo' | expect.to.start_with('fo')
        [1, 2, 3] | expect.to.start_with.number(1)
        iter([1, 2, 3]) | expect.to.start_with.numbers(1, 2)
        OrderedDict([('foo', 0), ('bar', 1)]) | expect.to.start_with('foo')

        # Expect style - negation form
        'foo' | expect.to_not.start_with('f')
        'foo' | expect.to_not.start_with('fo')
        [1, 2, 3] | expect.to_not.start_with.number(1)
        iter([1, 2, 3]) | expect.to_not.start_with.numbers(1, 2)
        OrderedDict([('foo', 0), ('bar', 1)]) | expect.to_not.start_with('foo')
    """

    # Operator keywords
    operators = ('start_with', 'starts_with')

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'an object that starts with items "{value}"',
        'an object that does not start with items "{value}"',
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with value "{value}"',
    )

    def matches(self, subject, *expected):
        if isinstance(subject, STR_TYPES):
            return (
                subject.startswith(expected[0]),
                ['starts with {0!r}'.format(subject[:-len(expected[0])])])

        head = list(subject)[:len(expected)]
        return (
            list(expected) == head,
            ['starts with {0!r}'.format(head)])


class EndWithOperator(StarEndBaseOpeartor):
    """
    Asserts if a given value ends with a specific items.

    Example::

        # Should style
        'foo' | should.ends_with('o')
        'foo' | should.ends_with('oo')
        [1, 2, 3] | should.ends_with.number(3)
        iter([1, 2, 3]) | should.ends_with.numbers(2, 3)
        OrderedDict([('foo', 0), ('bar', 1)]) | should.ends_with.item('bar')

        # Should style - negation form
        'foo' | should.do_not.ends_with('f')
        'foo' | should.do_not.ends_with('o')
        [1, 2, 3] | should.do_not.ends_with(2)
        iter([1, 2, 3]) | should.do_not.ends_with.numbers(3, 4)
        OrderedDict([('foo', 0), ('bar', 1)]) | should.ends_with('foo')

        # Expect style
        'foo' | expect.to.ends_with('o')
        'foo' | expect.to.ends_with('oo')
        [1, 2, 3] | expect.to.ends_with.number(3)
        iter([1, 2, 3]) | expect.to.ends_with.numbers(2, 3)
        OrderedDict([('foo', 0), ('bar', 1)]) | expect.to.ends_with('bar')

        # Expect style - negation form
        'foo' | expect.to_not.ends_with('f')
        'foo' | expect.to_not.ends_with('oo')
        [1, 2, 3] | expect.to_not.ends_with.number(2)
        iter([1, 2, 3]) | expect.to_not.ends_with.numbers(1, 2)
        OrderedDict([('foo', 0), ('bar', 1)]) | expect.to_not.ends_with('foo')
    """

    # Operator keywords
    operators = ('end_with', 'ends_with')

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'an object that ends with items "{value}"',
        'an object that does not end with items "{value}"',
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with value "{value}"',
    )

    def matches(self, subject, *expected):
        if isinstance(subject, STR_TYPES):
            return (
                subject.endswith(expected[0]),
                ['ends with {0!r}'.format(subject[-len(expected[0]):])])

        tail = list(subject)[-len(expected):]
        return (
            list(expected) == tail,
            ['ends with {0!r}'.format(tail)])



# -*- coding: utf-8 -*-
import functools


# List of unsupported binary operators
unsupported_operators = (
    # method    # operator symbol
    ('rshit',   '>>'),
    ('lshift',  '<<'),
    ('add',     '+'),
    ('and',     '&'),
)


def not_implemented(symbol, self, value):
    """
    Raises an `NotImplementedError` exception when called.
    """
    raise NotImplementedError('"{}" operator is not overloaded'.format(
        symbol
    ))


class BaseTest(object):
    """
    BaseTest implements magic method that explicitly raises an exception
    if consumed with unsupported binary operators.

    Note: methods are defined dynamically. See below.
    """


# Dynamically set class methods
for name, symbol in unsupported_operators:
    setattr(BaseTest, '__{}__'.format(name),
            functools.partial(not_implemented, symbol))



# -*- coding: utf-8 -*-
import types
import inspect
from array import array

from ..operator import Operator

# Type alias mappings
MAPPINGS = {
    'string': str,
    'int': int,
    'integer': int,
    'number': int,
    'object': object,
    'float': float,
    'bool': bool,
    'boolean': bool,
    'complex': complex,
    'list': list,
    'dict': dict,
    'dictionary': dict,
    'tuple': tuple,
    'set': set,
    'array': array,
    'lambda': types.LambdaType,
    'generator': types.GeneratorType,
    'asyncgenerator': getattr(types, 'GeneratorType', None),
    'class': 'class',
    'method': 'method',
    'module': 'module',
    'function': 'function',
    'coroutine': 'coroutine',
    'generatorfunction': 'generatorfunction',
    'generator function': 'generatorfunction',
    'coroutinefunction': 'coroutinefunction',
}


class TypeOperator(Operator):
    """
    Asserts if a given object satisfies a type.

    You can use both a type alias string or a ``type`` object.

    Example::

        # Should style
        1 | should.be.an('int')
        1 | should.be.an('number')
        True | should.be.a('bool')
        True | should.be.type(bool)
        'foo' | should.be.a(str)
        'foo' | should.be.a('string')
        [1, 2, 3] | should.be.a('list')
        [1, 2, 3] | should.have.type.of(list)
        (1, 2, 3) | should.be.a('tuple')
        (1, 2, 3) | should.have.type.of(tuple)
        (lamdba x: x) | should.be.a('lambda')
        'foo' | should.be.instance.of('str')

        # Should style - negation form
        1 | should.not_be.an('int')
        1 | should.not_be.an('number')
        True | should.not_be.a('bool')
        True | should.not_be.type(bool)
        'foo' | should.not_be.a(str)
        'foo' | should.not_be.a('string')
        [1, 2, 3] | should.not_be.a('list')
        [1, 2, 3] | should.have_not.type.of(list)
        (1, 2, 3) | should.not_be.a('tuple')
        (1, 2, 3) | should.have_not.type.of(tuple)
        (lamdba x: x) | should.not_be.a('lambda')

        # Expect style
        1 | expect.to.be.an('int')
        1 | expect.to.be.an('number')
        True | expect.to.be.a('bool')
        True | expect.to.be.type(bool)
        'foo' | expect.to.be.a(str)
        'foo' | expect.to.be.a('string')
        [1, 2, 3] | expect.to.be.a('list')
        [1, 2, 3] | expect.to.have.type.of(list)
        (1, 2, 3) | expect.to.be.a('tuple')
        (1, 2, 3) | expect.to.have.type.of(tuple)
        (lamdba x: x) | expect.to.be.a('lambda')

        # Expect style - negation form
        1 | expect.to_not.be.an('int')
        1 | expect.to_not.be.an('number')
        True | expect.to_not.be.a('bool')
        True | expect.to_not.be.type(bool)
        'foo' | expect.to_not.be.a(str)
        'foo' | expect.to_not.be.a('string')
        [1, 2, 3] | expect.to_not.be.a('list')
        [1, 2, 3] | expect.to_not.have.type.of(list)
        (1, 2, 3) | expect.to_not.be.a('tuple')
        (1, 2, 3) | expect.to_not.have.type.of(tuple)
        (lamdba x: x) | expect.to_not.be.a('lambda')
        'foo' | expect.to.be.instance.of('str')
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('type', 'types', 'a', 'an', 'instance')

    # Operator chain aliases
    aliases = ('type', 'types', 'of', 'equal', 'to')

    # Subject message template
    expected_message = Operator.Dsl.Message(
        'an object that is a "{value}" type',
        'an object that is not a "{value}" type'
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with value "{value}"'
    )

    def match(self, value, expected):
        # Custom expectations yielded values
        self.value = type(value).__name__
        self.expected = type(expected).__name__

        # Get type alias
        if type(expected) is str:
            self.expected = expected
            _expected = MAPPINGS.get(expected)

            if not _expected:
                raise ValueError('unsupported type alias: {}'.format(expected))

            if type(_expected) is str:
                return getattr(inspect, 'is{}'.format(expected))(value)

            expected = _expected

        # Check None type
        if expected is None:
            return value is None

        return isinstance(value, expected)



import pytest
from collections import OrderedDict


def test_start_end_with(should):
    # Strings
    'foo' | should.end_with.letter('o')
    'bar' | should.start_with.character('b')

    # Iterables
    [1, 2, 3] | should.start_with.number(1)
    [1, 2, 3] | should.end_with.number(2, 3)
    iter([1, 2, 3]) | should.start_with.numbers(1, 2)

    # Mappings
    OrderedDict([('foo', 0), ('bar', 1)]) | should.start_with('foo', 'bar')

    with pytest.raises(AssertionError):
        [1, 2, 3] | should.start_with.numbers(2, 3)

    with pytest.raises(AssertionError):
        'foo' | should.start_with.letter('o')



import pytest


def test_should_present(should):
    1 | should.be.present
    'foo' | should.be.present
    [1, 2, 3] | should.be.present
    (1, 2, 3) | should.be.present
    iter([1]) | should.be.present
    {'foo': True} | should.be.present

    '' | should.not_be.present
    0 | should.not_be.present
    None | should.not_be.present
    [] | should.not_be.present
    tuple() | should.not_be.present
    dict() | should.not_be.present

    with pytest.raises(AssertionError):
        'foo' | should.do_not.exists

    with pytest.raises(AssertionError):
        None | should.be.present

    with pytest.raises(AssertionError):
        None | should.exists

    with pytest.raises(AssertionError):
        0 | should.exists

    with pytest.raises(AssertionError):
        [] | should.exists

    with pytest.raises(AssertionError):
        tuple() | should.exists



import pytest


def test_raises(should):
    def error():
        raise AssertionError('foo')

    error | should.raise_error(AssertionError)
    error | should.do_not.raise_error(NotImplementedError)

    with pytest.raises(AssertionError):
        error | should.raise_error(NotImplementedError)

    with pytest.raises(AssertionError):
        error | should.do_not.raise_error(AssertionError)

    with pytest.raises(AssertionError):
        None | should.raise_error(AssertionError)



# -*- coding: utf-8 -*-
import sys
import logging
from colorama import Fore, Style

# Create logger instance for grappa
log = logging.getLogger('grappa')

# Custom log message formatter with colored output
formatter = logging.Formatter(
    u'{}=>{} {}[grappa]{} {}%(asctime)s{} | %(message)s'.format(
        Fore.GREEN,
        Style.RESET_ALL,
        Fore.MAGENTA,
        Style.RESET_ALL,
        Fore.CYAN,
        Style.RESET_ALL,
    )
)

# Create log handler with custom text formatter
handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(formatter)
log.addHandler(handler)

# Set default logging level
log.setLevel(logging.CRITICAL)



# -*- coding: utf-8 -*-
from ..operator import Operator


class NoneOperator(Operator):
    """
    Asserts if a given subject is ``None``.

    Example::

        # Should style
        None | should.be.none

        # Should style - negation form
        'foo' | should.not_be.none

        # Expect style
        None | expect.to.be.none

        # Expect style - negation form
        'foo' | expect.to_not.be.none
    """

    # Is the operator a keyword
    kind = Operator.Type.ACCESSOR

    # Operator keywords
    operators = ('none',)

    # Expected message templates
    expected_message = Operator.Dsl.Message(
        'a value that is exactly equal to "None"',
        'a value that is not equal to "None"',
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object with type "{type}" with value "{value}"',
    )

    # Assertion information
    information = (
        Operator.Dsl.Help(
            Operator.Dsl.Description(
                '"None" is a built-in constant in Python that represents the',
                'absence of a value, as when default arguments are not passed',
                'to a function. The sole value of the type NoneType.',
            ),
            Operator.Dsl.Reference(
                'https://docs.python.org/3/library/constants.html#None'
            ),
        ),
    )

    def match(self, subject):
        return subject is None, []



# -*- coding: utf-8 -*-
import re
import linecache
import traceback
import functools
from colorama import Fore, Style

from .empty import empty
from .config import config
from .operator_dsl import Message
from .template import ErrorTemplate


class BaseReporter(object):

    def __init__(self, ctx, error):
        self.ctx = ctx
        self.error = error

    def cut(self, value, size=50):
        text = str(value)
        return text[0:size] + ' ...' if len(text) > size else text

    def normalize(self, value, size=20):
        if value is None:
            return value

        if not hasattr(value, '__len__'):
            return value

        return self.cut(value)

    def safe_length(self, value):
        try:
            return len(value)
        except:
            return '"unmeasurable"'

    def from_operator(self, name, defaults=None):
        if not hasattr(self.error, 'operator'):
            return defaults
        value = getattr(self.error.operator, name, defaults)
        return defaults if value is empty else value

    def render_tmpl(self, tmpl, value):
        placeholders = {}

        if '{value}' in tmpl:
            placeholders['value'] = self.normalize(value)

        if '{type}' in tmpl:
            placeholders['type'] = type(value).__name__

        if '{length}' in tmpl:
            placeholders['length'] = self.safe_length(value)

        return tmpl.format(**placeholders)

    def run(self, error):
        raise NotImplementedError('run() method must be implemented')


class AssertionReporter(BaseReporter):

    title = 'The following assertion was not satisfied'

    template = 'subject "{}" {} {}'

    def run(self, error):
        # Assertion expression value
        subject = self.normalize(
            self.from_operator('subject', self.ctx.subject))

        # List of keyword operators DSL
        operators = ' '.join(self.ctx.keywords).replace('_', ' ')

        # Assertion expression value
        assertion = self.template.format(subject, self.ctx.style, operators)

        # Expected value
        expected = self.from_operator('expected', self.ctx.expected)
        if isinstance(expected, tuple):
            if len(expected) == 0:
                expected = empty
            if len(expected) == 1:
                expected = expected[0]

        # Add expected value template, if needed
        if expected is not empty:
            if isinstance(expected, (tuple, list)):
                expected = ', '.join(str(i) for i in expected)
            assertion += ' "{}"'.format(self.normalize(expected))

        return assertion


class SubjectMessageReporter(BaseReporter):

    title = 'What we got instead'

    # Attribute to lookup in the operator instance for custom message template
    attribute = 'subject'

    def run(self, error):
        if not hasattr(error, 'operator'):
            return None

        # Custom value human-friendly message
        value = getattr(self.ctx, self.attribute, None)

        # Value should not be empty
        if value is None or value is empty:
            value = self.from_operator(self.attribute, value)

        if value is None:
            return None if self.attribute == 'expected' else 'None'

        # Get first argument, if needed
        if self.attribute == 'expected' and len(value) == 1:
            value = value[0]

        # Get expectation message, if present in the operator
        attribute = '{}_message'.format(self.attribute)
        text_message = self.from_operator(attribute, None)
        if text_message:
            if isinstance(text_message, Message):
                attr = 'negation' if self.ctx.negate else 'allowance'
                text_message = getattr(text_message, attr, '')

            # Render template
            text_message = self.render_tmpl(text_message, value)

        return text_message or self.normalize(value)


class ExpectedMessageReporter(SubjectMessageReporter):

    title = 'What we expected'

    # Attribute to lookup in the operator instance for custom message template
    attribute = 'expected'


class MessageReporter(BaseReporter):

    title = 'Message'

    def run(self, error):
        return self.ctx.message


class ErrorReasonsReporter(BaseReporter):

    title = 'Reasons'

    def run(self, error):
        return getattr(error, 'reasons', None)


class InformationReporter(BaseReporter):

    title = 'Information'

    def run(self, error):
        return self.from_operator('information', None)


class UnexpectedError(BaseReporter):

    title = 'However, an unexpected exception was raised'

    def run(self, error):
        err = getattr(error, 'error', None)
        if not err:
            return None

        error.reasons = [
            'the assertion raised an unexpected exception, but it should not',
            'this behavior might be an bug within the testing library or '
            'in third-party test operators'
        ]

        return str(err)


class Trace(object):
    """
    Python < 3.4 traceback compatibility wrapper for Python +3.5
    """

    def __init__(self, trace):
        self.name = trace[2]
        self.line = trace[-1]
        self.lineno = trace[1]
        self.filename = trace[0]


class CodeReporter(BaseReporter):
    """
    CodeReporter matches and renders the fragment of the code
    """

    title = 'Where'

    LINES = 8
    INDENT_SPACES = 4

    PIPE_EXPR = re.compile(r'[\|][\s+]?(should|expect)[\.]')
    FN_CALL_EXPR = re.compile(r'(should|expect)[\s+]?[\(|\.]')

    def match_line(self, line):
        return any([
            CodeReporter.PIPE_EXPR.search(line),
            CodeReporter.FN_CALL_EXPR.match(line),
        ])

    def find_trace(self):
        # Get latest exception stack
        trace_list = list(traceback.extract_stack())
        trace_list.reverse()

        for trace in trace_list:
            # Normalize traceback in old Python versions
            if isinstance(trace, tuple):
                trace = Trace(trace)
            # Match code line
            if self.match_line(trace.line):
                return trace

    def header(self, trace):
        return 'File "{}", line {}, in {}\n\n'.format(
            trace.filename,
            trace.lineno,
            trace.name,
        )

    def render_code(self, trace):
        lmin = trace.lineno - CodeReporter.LINES
        lmax = trace.lineno + CodeReporter.LINES
        max_len = max(len(str(lmin)), len(str(lmax)))

        def calculate_space(line):
            return ' ' * (max_len - len(str(line)))

        def render_lines(buf, line):
            # Retrieve line of code from the trackback
            code = linecache.getline(trace.filename, line)
            if len(code) == 0:
                return buf

            # Calculate indentation and line head space for aligment
            space = calculate_space(line)
            indent = ' ' * CodeReporter.INDENT_SPACES

            # Line head
            head = '{}{}{}| {}'.format(
                Fore.GREEN if config.use_colors else '',
                space, line,
                Style.RESET_ALL if config.use_colors else ''
            )

            if line == trace.lineno:
                if config.use_colors:
                    head += Fore.RED + '>' + Style.RESET_ALL
                else:
                    head += '>'
            else:
                head += ' '

            buf.append(indent + head + code)
            return buf

        code = functools.reduce(render_lines, range(lmin, lmax), [])
        return self.header(trace) + (''.join(code))

    def run(self, error):
        if not config.show_code:
            return None

        trace = self.find_trace()
        return self.render_code(trace) if trace else None


class ErrorReporter(object):
    """
    ErrorReporter renders the current error code.
    """

    # Stores error message reporters
    # Reporters will be executed in definition order.
    REPORTERS = [
        AssertionReporter,
        MessageReporter,
        UnexpectedError,
        ErrorReasonsReporter,
        ExpectedMessageReporter,
        SubjectMessageReporter,
        InformationReporter,
        CodeReporter,
    ]

    def __init__(self, ctx):
        self.ctx = ctx

    def render_reporters(self, template, error):
        for Reporter in ErrorReporter.REPORTERS:
            report = Reporter(self.ctx, error).run(error)
            template.block(Reporter.title, report)

    def run(self, error):
        # Create error template generator
        template = ErrorTemplate()

        # Trigger registered reporters
        try:
            self.render_reporters(template, error)
        except Exception as err:
            err.__legit__ = True
            return err

        # Create assertion error
        err = AssertionError(template.render())

        # Flag error as grappa generated error
        err.__grappa__ = True
        err.error = error
        err.context = self.ctx

        return err



# -*- coding: utf-8 -*-


class GrappaAssertionError(AssertionError):
    """
    GrappaAssertionError represents an internal grappa assertion error
    and it is used internally to encapsulate and share state on
    assertion errors.

    Attributes:
        error (Exception): stores original error exception.
        reasons (list|tuple[str]): stores error reason details.
        operator (Operator): original operator instance that failed.
    """

    def __init__(self, error=None, reasons=None, operator=None):
        self.error = error
        self.reasons = reasons
        self.operator = operator



# -*- coding: utf-8 -*
"""
`grappa` provides two different testing styles: `should` and `expect`.

should
------

Example using ``should`` style::

    from grappa import should

    should('foo').be.equal.to('foo')
    'foo' | should.be.equal.to('foo')

expect
------

Example using ``expect`` style::

    from grappa import expect

    expect([1, 2, 3]).to.contain([2, 3])
    [1, 2, 3] | expect.to.contain([2, 3])


For assertion operators and aliases, see `operators documentation`_.

.. _`operators documentation`: operators.html

Reference
---------
"""

# Export public API module members
from .api import *  # noqa
from .api import __all__  # noqa

# Package metadata
__author__ = 'Tomas Aparicio'
__license__ = 'MIT'

# Current package version
__version__ = '0.1.0'



# -*- coding: utf-8 -*-
from colorama import Fore, Style


class Message(object):

    def __init__(self, allowance=None, negation=None):
        self.allowance = allowance
        self.negation = negation or allowance

    def render(self, negation=False):
        return self.negation if negation else self.allowance


class Description(object):

    ARROW = '>'

    def __init__(self, *text):
        self.value = text

    def render(self, separator):
        lines = '\n{}'.format(separator).join(self.value)
        return Description.ARROW + ' ' + lines


class Reference(object):

    TOKEN = '=>'
    TEXT = 'Reference'

    def __init__(self, url=None):
        self.url = url

    def render(self, separator):
        return separator + '{}{}{} {}{}: {}{}'.format(
            Fore.CYAN,
            Reference.TOKEN,
            Style.RESET_ALL,
            Fore.GREEN,
            Reference.TEXT,
            self.url,
            Style.RESET_ALL,
        )


class Help(object):

    def __init__(self, description=None, *references):
        self.description = description
        self.references = references

    def render(self, separator):
        description = self.description.render(separator)
        references = [ref.render(separator) for ref in self.references]
        return '\n'.join([description] + references)



import pytest
from grappa.operators.bool import TrueOperator, FalseOperator


def test_should_true(should):
    True | should.be.true

    with pytest.raises(AssertionError):
        False | should.be.true

    with pytest.raises(AssertionError):
        None | should.be.true

    with pytest.raises(AssertionError):
        'foo' | should.be.true

    with pytest.raises(AssertionError):
        [1, 2, 3] | should.be.true


def test_expect_true(expect):
    True | expect.to.be.true

    with pytest.raises(AssertionError):
        False | expect.to.be.true

    with pytest.raises(AssertionError):
        None | expect.to.be.true

    with pytest.raises(AssertionError):
        'foo' | expect.to.be.true

    with pytest.raises(AssertionError):
        [1, 2, 3] | expect.to.be.true


def test_should_false(should):
    False | should.be.false

    with pytest.raises(AssertionError):
        True | should.be.false

    with pytest.raises(AssertionError):
        None | should.be.false

    with pytest.raises(AssertionError):
        'foo' | should.be.false

    with pytest.raises(AssertionError):
        [1, 2, 3] | should.be.false


def test_expect_false(expect):
    False | expect.to.be.false

    with pytest.raises(AssertionError):
        True | expect.to.be.false

    with pytest.raises(AssertionError):
        None | expect.to.be.false

    with pytest.raises(AssertionError):
        'foo' | expect.to.be.false

    with pytest.raises(AssertionError):
        [1, 2, 3] | expect.to.be.false


def test_true_operator(ctx):
    assert TrueOperator(ctx).match(True) == (True, [])
    assert TrueOperator(ctx).match(False) == (False, [])

    assert TrueOperator(ctx).match(0) == (False,
                                          ['subject is not a bool type'])


def test_true_operator_message(should):
    (TrueOperator
        | should.have.property('kind')
        > should.be.equal.to('accessor'))

    (TrueOperator
        | should.have.property('operators')
        > should.be.equal.to(('true',)))

    TrueOperator | should.have.property('aliases') > should.be.empty

    TrueOperator | should.have.property('expected_message')


def test_false_operator(ctx):
    assert FalseOperator(ctx).match(False) == (True, [])
    assert FalseOperator(ctx).match(True) == (False, [])

    assert FalseOperator(ctx).match(0) == (False,
                                           ['subject is not a bool type'])


def test_false_operator_message(should):
    (FalseOperator
        | should.have.property('kind')
        > should.be.equal.to('accessor'))

    (FalseOperator
        | should.have.property('operators')
        > should.be.equal.to(('false',)))

    FalseOperator | should.have.property('aliases') > should.be.empty

    FalseOperator | should.have.property('expected_message')



import re
import pytest


def test_expect_match(should):
    'Hello foo bar' | should.match(r'Hello \w+ bar')
    'Hello foo BAR' | should.match.regexp(r'Hello \w+ bar', re.I)
    'Hello foo BAR' | should.match.regexp('BAR', re.I)

    with pytest.raises(AssertionError):
        should('foo bar').match('baz')

    with pytest.raises(AssertionError):
        'foo bar' | should.match.value('baz')

    with pytest.raises(AssertionError):
        'Hello foo BAR' | should.match.regexp(r'Hello \w+ bar')



# -*- coding: utf-8 -*-
import inspect
import functools
from ..operator import Operator


class ImplementsOperator(Operator):
    """
    Asserts if a given object implements an interface of methods.

    Example::

        class Foo(object):
            def bar():
                pass

            def baz():
                pass

        # Should style
        Foo() | should.implements('bar')
        Foo() | should.implements.method('bar')
        Foo() | should.implement.methods('bar', 'baz')
        Foo() | should.implement.interface('bar', 'baz')
        Foo() | should.satisfies.interface('bar', 'baz')

        # Should style - negation form
        Foo() | should.do_not.implements('bar')
        Foo() | should.do_not.implement.methods('bar', 'baz')
        Foo() | should.do_not.implement.interface('bar', 'baz')
        Foo() | should.do_not.satisfy.interface('bar', 'baz')

        # Expect style
        Foo() | expect.to.implement('bar')
        Foo() | expect.to.implement.method('bar')
        Foo() | expect.to.implement.methods('bar', 'baz')
        Foo() | expect.to.implement.interface('bar', 'baz')
        Foo() | expect.to.satisfy.interface('bar', 'baz')

        # Expect style - negation form
        Foo() | expect.to_not.implement('bar')
        Foo() | expect.to_not.implement.method('bar')
        Foo() | expect.to_not.implement.methods('bar', 'baz')
        Foo() | expect.to_not.implement.interface('bar', 'baz')
        Foo() | expect.to_not.satisfy.interface('bar', 'baz')
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('implements', 'implement', 'satisfies', 'satisfy')

    # Operator chain aliases
    aliases = ('interface', 'methods', 'method')

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'an object that implements the following members "{value}"',
        'an object that does not implement the following members "{value}"',
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with content "{value}"',
    )

    # Assertion information
    information = (
        Operator.Dsl.Help(
            Operator.Dsl.Description(
                'Object interface implementation is verified by using the',
                'Python built-in function "hasattr" along with',
                '"inspect.ismethod()" function in order to infer if a given',
                'object implements the required methods.'
            ),
            Operator.Dsl.Reference(
                'https://docs.python.org/3.6/library/functions.html#hasattr'
            ),
        ),
    )

    def has_method(self, cls, method):
        return any([inspect.ismethod(getattr(cls, method)),
                    inspect.isfunction(getattr(cls, method))])

    def match(self, cls, *methods):
        if len(methods) == 0:
            return False, ['methods to implement cannot be empty']

        def validate(reasons, method):
            if not hasattr(cls, method):
                reasons.append('object does not have property "{}"'.format(
                    method
                ))
            elif not self.has_method(cls, method):
                reasons.append('object does not implement method "{}"'.format(
                    method
                ))
            return reasons

        reasons = functools.reduce(validate, methods, [])
        return len(reasons) == 0, reasons



import pytest


def test_implements(should):
    class foo(object):
        foo = None

        def bar():
            pass

        def baz():
            pass

    foo | should.implements.methods('bar', 'baz')
    foo | should.do_not.implements.methods('foo')

    with pytest.raises(AssertionError):
        foo | should.implement.methods('foo', 'faa')



# -*- coding: utf-8 -*-
from ..operator import Operator


class BelowOperator(Operator):
    """
    Asserts if a given number is below to another number.

    Example::

        # Should style
        3 | should.be.below(5)
        3 | should.be.below.of(5)
        3 | should.be.below.to(5)
        3 | should.be.less.than(5)
        3 | should.be.lower.than(5)
        3 | should.be.below.to.number(5)
        3 | should.be.below.than.number(5)

        # Should style - negation form
        5 | should.not_be.below(3)
        5 | should.not_be.below.of(3)
        3 | should.not_be.below.to(5)
        3 | should.not_be.lower.than(5)
        5 | should.not_be.below.to.number(3)

        # Expect style
        3 | expect.to.be.below(5)
        3 | expect.to.be.below.of(5)
        3 | expect.to.be.below.to(5)
        3 | expect.to.be.less.than(5)
        3 | expect.to.be.lower.than(5)
        3 | expect.to.be.below.to.number(5)
        3 | expect.to.be.below.than.number(5)

        # Expect style - negation form
        5 | expect.to_not.be.below(3)
        5 | expect.to_not.be.below.of(3)
        5 | expect.to_not.be.below.than(3)
        5 | expect.to_not.be.below.to.number(3)
        5 | expect.to_not.be.below.than.number(3)
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('below', 'lower', 'less')

    # Operator keywords
    aliases = ('of', 'to', 'number', 'than')

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'a number that is below to {value}',
        'a number that is not below to {value}'
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with value "{value}"',
    )

    def match(self, subject, expected):
        if self.ctx.length:
            subject = len(subject)

        if not isinstance(subject, (int, float, complex)):
            return False, ['subject is not a numeric type']

        return subject < expected, ['subject is not lower than {expected}']


class AboveOperator(BelowOperator):
    """
    Asserts if a given number is above to another number.

    Example::

        # Should style
        5 | should.be.above(3)
        5 | should.be.above.of(3)
        5 | should.be.above.to(3)
        5 | should.be.higher.than(3)
        5 | should.be.above.to.number(3)
        5 | should.be.above.than.number(3)

        # Should style - negation form
        3 | should.not_be.above(5)
        3 | should.not_be.above.of(5)
        3 | should.not_be.above.to(5)
        3 | should.not_be.higher.than(5)
        3 | should.not_be.above.to.number(5)
        3 | should.not_be.above.than.number(5)

        # Expect style
        5 | expect.to.be.above(3)
        5 | expect.to.be.above.of(3)
        5 | expect.to.be.above.to(3)
        5 | expect.to.be.higher.than(3)
        5 | expect.to.be.above.to.number(3)
        5 | expect.to.be.above.than.number(3)

        # Expect style - negation form
        3 | expect.not_to.be.above(5)
        3 | expect.not_to.be.above.of(5)
        3 | expect.not_to.be.above.to(5)
        3 | expect.not_to.be.higher.than(5)
        3 | expect.not_to.be.above.to.number(5)
        3 | expect.not_to.be.above.than.number(5)
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('above', 'higher')

    # Expected
    expected = 'a "True" boolean value'

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'a number that is above to {value}',
        'a number that is not above to {value}'
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with value "{value}"',
    )

    def match(self, subject, expected):
        if self.ctx.length:
            subject = len(subject)

        if not isinstance(subject, (int, float, complex)):
            return False, ['subject is not a numeric type']

        return subject > expected, []


class AboveOrEqualOperator(BelowOperator):
    """
    Asserts if a given number is above or equal to another number.

    Example::

        # Should style
        3 | should.be.least(3)
        3 | should.be.above_or_equal(3)
        3 | should.be.above_or_equal.of(3)
        3 | should.be.above_or_equal.to(3)
        3 | should.be.higher_or_equal.than(3)
        3 | should.be.above_or_equal.to.number(3)
        3 | should.be.above_or_equal.than.number(3)

        # Should style - negation form
        3 | should.not_be.least(3)
        3 | should.not_be.above_or_equal(5)
        3 | should.not_be.above_or_equal.of(5)
        3 | should.not_be.above_or_equal.to(5)
        3 | should.not_be.higher_or_equal.than(5)
        3 | should.not_be.higher_or_equal.to.number(5)
        3 | should.not_be.higher_or_equal.than.number(5)

        # Expect style
        3 | expect.to.be.least(3)
        3 | expect.to.be.above_or_equal(3)
        3 | expect.to.be.above_or_equal.of(3)
        3 | expect.to.be.above_or_equal.to(3)
        3 | expect.to.be.higher_or_equal.than(3)
        3 | expect.to.be.above_or_equal.to.number(3)
        3 | expect.to.be.above_or_equal.than.number(3)

        # Expect style - negation form
        3 | expect.not_be.least(3)
        3 | expect.not_be.above_or_equal(5)
        3 | expect.not_be.above_or_equal.of(5)
        3 | expect.not_be.above_or_equal.to(5)
        3 | expect.not_be.higher_or_equal.than(5)
        3 | expect.not_be.higher_or_equal.to.number(5)
        3 | expect.not_be.higher_or_equal.than.number(5)
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('above_or_equal', 'higher_or_equal', 'least')

    # Expected
    expected = 'a "True" boolean value'

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'a number that is higher or equal to {value}',
        'a number that is not higher or equal to {value}'
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with value "{value}"',
    )

    def match(self, subject, expected):
        if self.ctx.length:
            subject = len(subject)

        if not isinstance(subject, (int, float, complex)):
            return False, ['subject is not a numeric type']

        return subject >= expected, []


class BelowOrEqualOperator(BelowOperator):
    """
    Asserts if a given number is below or equal to another number.

    Example::

        # Should style
        3 | should.be.most(3)
        3 | should.be.below_or_equal(3)
        3 | should.be.below_or_equal.of(3)
        3 | should.be.below_or_equal.to(3)
        3 | should.be.lower_or_equal.than(3)
        3 | should.be.lower_or_equal.to.number(3)
        3 | should.be.lower_or_equal.than.number(3)

        # Should style - negation form
        3 | should.not_be.most(5)
        3 | should.not_be.below_or_equal(5)
        3 | should.not_be.below_or_equal.of(5)
        3 | should.not_be.below_or_equal.to(5)
        3 | should.not_be.lower_or_equal.than(5)
        3 | should.not_be.lower_or_equal.to.number(5)
        3 | should.not_be.lower_or_equal.than.number(5)

        # Expect style
        3 | expect.to.be.most(3)
        3 | expect.to.be.below_or_equal(3)
        3 | expect.to.be.below_or_equal.of(3)
        3 | expect.to.be.below_or_equal.to(3)
        3 | expect.to.be.lower_or_equal.than(3)
        3 | expect.to.be.lower_or_equal.to.number(3)
        3 | expect.to.be.lower_or_equal.than.number(3)

        # Expect style - negation form
        3 | expect.not_be.most(5)
        3 | expect.not_be.below_or_equal(5)
        3 | expect.not_be.below_or_equal.of(5)
        3 | expect.not_be.below_or_equal.to(5)
        3 | expect.not_be.lower_or_equal.than(5)
        3 | expect.not_be.lower_or_equal.to.number(5)
        3 | expect.not_be.lower_or_equal.than.number(5)
    """

    # Is the operator a keyword
    kind = Operator.Type.MATCHER

    # Operator keywords
    operators = ('below_or_equal', 'lower_or_equal', 'most')

    # Expected
    expected = 'a "True" boolean value'

    # Expected template message
    expected_message = Operator.Dsl.Message(
        'a number that is below or equal to {value}',
        'a number that is not below or equal to {value}'
    )

    # Subject template message
    subject_message = Operator.Dsl.Message(
        'an object of type "{type}" with value "{value}"',
    )

    def match(self, subject, expected):
        if self.ctx.length:
            subject = len(subject)

        if not isinstance(subject, (int, float, complex)):
            return False, ['subject is not a numeric type']

        return subject <= expected, []
